@@ -1,5 +1,5 @@
 /******************************************************************************
- * $Id: PJ_aea.c,v 1.3 2002/12/14 19:27:06 warmerda Exp $
+ * $Id$
  *
  * Project:  PROJ.4
  * Purpose:  Implementation of the aea (Albers Equal Area) projection.
@@ -25,13 +25,7 @@
  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  * DEALINGS IN THE SOFTWARE.
- ******************************************************************************
- *
- * $Log: PJ_aea.c,v $
- * Revision 1.3  2002/12/14 19:27:06  warmerda
- * updated header
- *
- */
+ *****************************************************************************/
 
 #define PROJ_PARMS__ \
 	double	ec; \
@@ -49,7 +43,7 @@
 #define PJ_LIB__
 #include <projects.h>
 
-PJ_CVSID("$Id: PJ_aea.c,v 1.3 2002/12/14 19:27:06 warmerda Exp $");
+PJ_CVSID("$Id$");
 
 # define EPS10	1.e-10
 # define TOL7	1.e-7
@@ -158,11 +152,11 @@ setup(PJ *P) {
 	P->inv = e_inverse; P->fwd = e_forward;
 	return P;
 }
-ENTRY0(aea)
-	P->phi1 = pj_param(P->params, "rlat_1").f;
-	P->phi2 = pj_param(P->params, "rlat_2").f;
+ENTRY1(aea,en)
+	P->phi1 = pj_param(P->ctx, P->params, "rlat_1").f;
+	P->phi2 = pj_param(P->ctx, P->params, "rlat_2").f;
 ENDENTRY(setup(P))
-ENTRY0(leac)
-	P->phi2 = pj_param(P->params, "rlat_1").f;
-	P->phi1 = pj_param(P->params, "bsouth").i ? - HALFPI: HALFPI;
+ENTRY1(leac,en)
+	P->phi2 = pj_param(P->ctx, P->params, "rlat_1").f;
+	P->phi1 = pj_param(P->ctx, P->params, "bsouth").i ? - HALFPI: HALFPI;
 ENDENTRY(setup(P))
