@@ -1,5 +1,5 @@
 /******************************************************************************
- * $Id: cs2cs.c,v 1.7 2003/03/25 17:53:13 warmerda Exp $
+ * $Id$
  *
  * Project:  PROJ.4
  * Purpose:  Mainline program sort of like ``proj'' for converting between
@@ -26,41 +26,17 @@
  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  * DEALINGS IN THE SOFTWARE.
- ******************************************************************************
- *
- * $Log: cs2cs.c,v $
- * Revision 1.7  2003/03/25 17:53:13  warmerda
- * modified so that -f formats are used for Z as well
- *
- * Revision 1.6  2002/12/09 16:01:02  warmerda
- * added prime meridian support
- *
- * Revision 1.5  2002/11/19 20:33:42  warmerda
- * cleanup before exit to facilitate memory leak testing
- *
- * Revision 1.4  2001/04/05 19:32:19  warmerda
- * use projPJ, and pj_is_latlong()
- *
- * Revision 1.3  2001/04/05 04:23:28  warmerda
- * use pj_latlong_from_proj
- *
- * Revision 1.2  2001/02/03 18:36:55  warmerda
- * removed some unavailable options from usage string
- *
- * Revision 1.1  2000/07/06 23:32:27  warmerda
- * New
- *
- */
+ *****************************************************************************/
 
+#include "projects.h"
 #include <stdio.h>
 #include <stdlib.h>
 #include <ctype.h>
 #include <string.h>
 #include <math.h>
-#include "projects.h"
 #include "emess.h"
 
-#define MAX_LINE 200
+#define MAX_LINE 1000
 #define MAX_PARGS 100
 
 static projPJ   fromProj, toProj;
@@ -140,6 +116,7 @@ static void process(FILE *fid)
             {
                 data.u = HUGE_VAL;
                 data.v = HUGE_VAL;
+                emess(-3,"pj_transform(): %s", pj_strerrno(pj_errno));
             }
         }
 
@@ -176,7 +153,10 @@ static void process(FILE *fid)
             printf( oform, z );
         else
             printf( "%.3f", z );
-        fputs("\n", stdout );
+        if( s )
+            printf( "%s", s );
+        else
+            printf( "\n" );
     }
 }
 
@@ -186,18 +166,17 @@ static void process(FILE *fid)
 
 int main(int argc, char **argv) 
 {
-    char *arg, **eargv = argv, *from_argv[MAX_PARGS], *to_argv[MAX_PARGS],
-        **iargv = argv;
+    char *arg, **eargv = argv, *from_argv[MAX_PARGS], *to_argv[MAX_PARGS];
     FILE *fid;
-    int from_argc=0, to_argc=0, iargc = argc, eargc = 0, c, mon = 0;
+    int from_argc=0, to_argc=0, eargc = 0, c, mon = 0;
     int have_to_flag = 0, inverse = 0, i;
 
-    if (emess_dat.Prog_name = strrchr(*argv,DIR_CHAR))
+    if ((emess_dat.Prog_name = strrchr(*argv,DIR_CHAR)) != NULL)
         ++emess_dat.Prog_name;
     else emess_dat.Prog_name = *argv;
     inverse = ! strncmp(emess_dat.Prog_name, "inv", 3);
     if (argc <= 1 ) {
-        (void)fprintf(stderr, usage, pj_release, emess_dat.Prog_name);
+        (void)fprintf(stderr, usage, pj_get_release(), emess_dat.Prog_name);
         exit (0);
     }
     /* process run line arguments */
@@ -227,7 +206,7 @@ int main(int argc, char **argv)
                     int do_long = arg[1] == 'P', c;
                     char *str;
 
-                    for (lp = pj_list ; lp->id ; ++lp) {
+                    for (lp = pj_get_list_ref() ; lp->id ; ++lp) {
                         (void)printf("%s : ", lp->id);
                         if (do_long)  /* possibly multiline description */
                             (void)puts(*lp->descr);
@@ -242,7 +221,7 @@ int main(int argc, char **argv)
                     struct PJ_LIST *lp;
 
                     arg += 2;
-                    for (lp = pj_list ; lp->id ; ++lp)
+                    for (lp = pj_get_list_ref() ; lp->id ; ++lp)
                         if (!strcmp(lp->id, arg)) {
                             (void)printf("%9s : %s\n", lp->id, *lp->descr);
                             break;
@@ -250,20 +229,20 @@ int main(int argc, char **argv)
                 } else if (arg[1] == 'e') { /* list ellipses */
                     struct PJ_ELLPS *le;
 
-                    for (le = pj_ellps; le->id ; ++le)
+                    for (le = pj_get_ellps_ref(); le->id ; ++le)
                         (void)printf("%9s %-16s %-16s %s\n",
                                      le->id, le->major, le->ell, le->name);
                 } else if (arg[1] == 'u') { /* list units */
                     struct PJ_UNITS *lu;
 
-                    for (lu = pj_units; lu->id ; ++lu)
+                    for (lu = pj_get_units_ref(); lu->id ; ++lu)
                         (void)printf("%12s %-20s %s\n",
                                      lu->id, lu->to_meter, lu->name);
                 } else if (arg[1] == 'd') { /* list datums */
                     struct PJ_DATUMS *ld;
 
                     printf("__datum_id__ __ellipse___ __definition/comments______________________________\n" );
-                    for (ld = pj_datums; ld->id ; ++ld)
+                    for (ld = pj_get_datums_ref(); ld->id ; ++ld)
                     {
                         printf("%12s %-12s %-30s\n",
                                ld->id, ld->ellipse_id, ld->defn);
@@ -273,7 +252,7 @@ int main(int argc, char **argv)
                 } else if( arg[1] == 'm') { /* list prime meridians */
                     struct PJ_PRIME_MERIDIANS *lpm;
 
-                    for (lpm = pj_prime_meridians; lpm->id ; ++lpm)
+                    for (lpm = pj_get_prime_meridians_ref(); lpm->id ; ++lpm)
                         (void)printf("%12s %-30s\n",
                                      lpm->id, lpm->defn);
                 } else
@@ -304,6 +283,10 @@ int main(int argc, char **argv)
               case 's': /* reverse output */
                 reverseout = 1;
                 continue;
+              case 'd': /* set debug level */
+                if (--argc <= 0) goto noargument;
+                pj_ctx_set_debug( pj_get_default_ctx(), atoi(*++argv));
+                continue;
               default:
                 emess(1, "invalid option: -%c",*arg);
                 break;
@@ -356,7 +339,11 @@ int main(int argc, char **argv)
         to_argc = argcount;
     }
 
-    if (!(fromProj = pj_init(from_argc, from_argv)))
+    if( from_argc == 0 && to_argc != 0 )
+    {
+        /* we will generate the from proj as the latlong of the +to in a bit */
+    }
+    else if (!(fromProj = pj_init(from_argc, from_argv)))
     {
         printf( "Using from definition: " );
         for( i = 0; i < from_argc; i++ )
@@ -391,6 +378,20 @@ int main(int argc, char **argv)
               pj_strerrno(pj_errno));
     }
 
+    if( from_argc == 0 && toProj != NULL) 
+    {
+        if (!(fromProj = pj_latlong_from_proj( toProj )))
+        {
+            printf( "Using to definition: " );
+            for( i = 0; i < to_argc; i++ )
+                printf( "%s ", to_argv[i] );
+            printf( "\n" );
+            
+            emess(3,"projection initialization failure\ncause: %s",
+                  pj_strerrno(pj_errno));
+        }   
+    }
+
     if (mon) {
         printf( "%c ---- From Coordinate System ----\n", tag );
         pj_pr_list(fromProj);
