@@ -1,5 +1,5 @@
 /******************************************************************************
- * $Id: pj_gridinfo.c,v 1.4 2003/03/19 03:36:41 warmerda Exp $
+ * $Id$
  *
  * Project:  PROJ.4
  * Purpose:  Functions for handling individual PJ_GRIDINFO's.  Includes
@@ -26,22 +26,7 @@
  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  * DEALINGS IN THE SOFTWARE.
- ******************************************************************************
- *
- * $Log: pj_gridinfo.c,v $
- * Revision 1.4  2003/03/19 03:36:41  warmerda
- * Fixed so swap_words() works when it should.
- *
- * Revision 1.3  2003/03/17 19:44:45  warmerda
- * improved debugging, reduce header read size
- *
- * Revision 1.2  2003/03/17 18:56:34  warmerda
- * implement heirarchical NTv2 gridinfos
- *
- * Revision 1.1  2003/03/15 06:01:18  warmerda
- * New
- *
- */
+ *****************************************************************************/
 
 #define PJ_LIB__
 
@@ -49,7 +34,16 @@
 #include <string.h>
 #include <math.h>
 #include <errno.h>
-#include <assert.h>
+
+#ifdef _WIN32_WCE
+/* assert.h includes all Windows API headers and causes 'LP' name clash.
+ * Here assert we disable assert() for Windows CE.
+ * TODO - mloskot: re-implement porting friendly assert
+ */
+# define assert(exp)	((void)0)
+#else
+# include <assert.h>
+#endif /* _WIN32_WCE */
 
 /************************************************************************/
 /*                             swap_words()                             */
@@ -68,16 +62,16 @@ static void swap_words( unsigned char *data, int word_size, int word_count )
     for( word = 0; word < word_count; word++ )
     {
         int	i;
-        
+
         for( i = 0; i < word_size/2; i++ )
         {
             int	t;
-            
+
             t = data[i];
             data[i] = data[word_size-i-1];
             data[word_size-i-1] = t;
         }
-        
+
         data += word_size;
     }
 }
@@ -86,7 +80,7 @@ static void swap_words( unsigned char *data, int word_size, int word_count )
 /*                          pj_gridinfo_free()                          */
 /************************************************************************/
 
-void pj_gridinfo_free( PJ_GRIDINFO *gi )
+void pj_gridinfo_free( projCtx ctx, PJ_GRIDINFO *gi )
 
 {
     if( gi == NULL )
@@ -99,13 +93,13 @@ void pj_gridinfo_free( PJ_GRIDINFO *gi )
         for( child = gi->child; child != NULL; child=next)
         {
             next=child->next;
-            pj_gridinfo_free( child );
+            pj_gridinfo_free( ctx, child );
         }
     }
 
     if( gi->ct != NULL )
         nad_free( gi->ct );
-    
+
     free( gi->gridname );
     if( gi->filename != NULL )
         free( gi->filename );
@@ -121,33 +115,74 @@ void pj_gridinfo_free( PJ_GRIDINFO *gi )
 /*      stuff are loaded by pj_gridinfo_init().                         */
 /************************************************************************/
 
-int pj_gridinfo_load( PJ_GRIDINFO *gi )
+int pj_gridinfo_load( projCtx ctx, PJ_GRIDINFO *gi )
 
 {
+    struct CTABLE ct_tmp;
+
     if( gi == NULL || gi->ct == NULL )
         return 0;
 
+    pj_acquire_lock();
+    if( gi->ct->cvs != NULL )
+    {
+        pj_release_lock();
+        return 1;
+    }
+
+    memcpy(&ct_tmp, gi->ct, sizeof(struct CTABLE));
+
 /* -------------------------------------------------------------------- */
-/*      ctable is currently loaded on initialization though there is    */
-/*      no real reason not to support delayed loading for it as well.   */
+/*      Original platform specific CTable format.                       */
 /* -------------------------------------------------------------------- */
     if( strcmp(gi->format,"ctable") == 0 )
     {
-        FILE *fid;
+        PAFile fid;
         int result;
 
-        fid = pj_open_lib( gi->filename, "rb" );
-        
+        fid = pj_open_lib( ctx, gi->filename, "rb" );
+
         if( fid == NULL )
         {
-            pj_errno = -38;
+            pj_ctx_set_errno( ctx, -38 );
+            pj_release_lock();
             return 0;
         }
 
-        result = nad_ctable_load( gi->ct, fid );
+        result = nad_ctable_load( ctx, &ct_tmp, fid );
 
-        fclose( fid );
+        pj_ctx_fclose( ctx, fid );
 
+        gi->ct->cvs = ct_tmp.cvs;
+        pj_release_lock();
+
+        return result;
+    }
+
+/* -------------------------------------------------------------------- */
+/*      CTable2 format.                                                 */
+/* -------------------------------------------------------------------- */
+    else if( strcmp(gi->format,"ctable2") == 0 )
+    {
+        PAFile fid;
+        int result;
+
+        fid = pj_open_lib( ctx, gi->filename, "rb" );
+
+        if( fid == NULL )
+        {
+            pj_ctx_set_errno( ctx, -38 );
+            pj_release_lock();
+            return 0;
+        }
+
+        result = nad_ctable2_load( ctx, &ct_tmp, fid );
+
+        pj_ctx_fclose( ctx, fid );
+
+        gi->ct->cvs = ct_tmp.cvs;
+
+        pj_release_lock();
         return result;
     }
 
@@ -162,38 +197,41 @@ int pj_gridinfo_load( PJ_GRIDINFO *gi )
     {
         double	*row_buf;
         int	row;
-        FILE *fid;
+        PAFile fid;
+
+        fid = pj_open_lib( ctx, gi->filename, "rb" );
 
-        fid = pj_open_lib( gi->filename, "rb" );
-        
         if( fid == NULL )
         {
-            pj_errno = -38;
+            pj_ctx_set_errno( ctx, -38 );
+            pj_release_lock();
             return 0;
         }
 
-        fseek( fid, gi->grid_offset, SEEK_SET );
+        pj_ctx_fseek( ctx, fid, gi->grid_offset, SEEK_SET );
 
         row_buf = (double *) pj_malloc(gi->ct->lim.lam * sizeof(double) * 2);
-        gi->ct->cvs = (FLP *) pj_malloc(gi->ct->lim.lam*gi->ct->lim.phi*sizeof(FLP));
-        if( row_buf == NULL || gi->ct->cvs == NULL )
+        ct_tmp.cvs = (FLP *) pj_malloc(gi->ct->lim.lam*gi->ct->lim.phi*sizeof(FLP));
+        if( row_buf == NULL || ct_tmp.cvs == NULL )
         {
-            pj_errno = -38;
+            pj_ctx_set_errno( ctx, -38 );
+            pj_release_lock();
             return 0;
         }
-        
+
         for( row = 0; row < gi->ct->lim.phi; row++ )
         {
             int	    i;
             FLP     *cvs;
             double  *diff_seconds;
 
-            if( fread( row_buf, sizeof(double), gi->ct->lim.lam * 2, fid ) 
+            if( pj_ctx_fread( ctx, row_buf,
+                              sizeof(double), gi->ct->lim.lam * 2, fid )
                 != 2 * gi->ct->lim.lam )
             {
                 pj_dalloc( row_buf );
-                pj_dalloc( gi->ct->cvs );
-                pj_errno = -38;
+                pj_dalloc( ct_tmp.cvs );
+                pj_ctx_set_errno( ctx, -38 );
                 return 0;
             }
 
@@ -205,7 +243,7 @@ int pj_gridinfo_load( PJ_GRIDINFO *gi )
 
             for( i = 0; i < gi->ct->lim.lam; i++ )
             {
-                cvs = gi->ct->cvs + (row) * gi->ct->lim.lam
+                cvs = ct_tmp.cvs + (row) * gi->ct->lim.lam
                     + (gi->ct->lim.lam - i - 1);
 
                 cvs->phi = *(diff_seconds++) * ((PI/180.0) / 3600.0);
@@ -215,7 +253,10 @@ int pj_gridinfo_load( PJ_GRIDINFO *gi )
 
         pj_dalloc( row_buf );
 
-        fclose( fid );
+        pj_ctx_fclose( ctx, fid );
+
+        gi->ct->cvs = ct_tmp.cvs;
+        pj_release_lock();
 
         return 1;
     }
@@ -231,49 +272,50 @@ int pj_gridinfo_load( PJ_GRIDINFO *gi )
     {
         float	*row_buf;
         int	row;
-        FILE *fid;
+        PAFile fid;
 
-        if( getenv("PROJ_DEBUG") != NULL )
-        {
-            fprintf( stderr, "NTv2 - loading grid %s\n", gi->ct->id );
-        }
+        pj_log( ctx, PJ_LOG_DEBUG_MINOR,
+                "NTv2 - loading grid %s", gi->ct->id );
+
+        fid = pj_open_lib( ctx, gi->filename, "rb" );
 
-        fid = pj_open_lib( gi->filename, "rb" );
-        
         if( fid == NULL )
         {
-            pj_errno = -38;
+            pj_ctx_set_errno( ctx, -38 );
+            pj_release_lock();
             return 0;
         }
 
-        fseek( fid, gi->grid_offset, SEEK_SET );
+        pj_ctx_fseek( ctx, fid, gi->grid_offset, SEEK_SET );
 
         row_buf = (float *) pj_malloc(gi->ct->lim.lam * sizeof(float) * 4);
-        gi->ct->cvs = (FLP *) pj_malloc(gi->ct->lim.lam*gi->ct->lim.phi*sizeof(FLP));
-        if( row_buf == NULL || gi->ct->cvs == NULL )
+        ct_tmp.cvs = (FLP *) pj_malloc(gi->ct->lim.lam*gi->ct->lim.phi*sizeof(FLP));
+        if( row_buf == NULL || ct_tmp.cvs == NULL )
         {
-            pj_errno = -38;
+            pj_ctx_set_errno( ctx, -38 );
+            pj_release_lock();
             return 0;
         }
-        
+
         for( row = 0; row < gi->ct->lim.phi; row++ )
         {
             int	    i;
             FLP     *cvs;
             float   *diff_seconds;
 
-            if( fread( row_buf, sizeof(float), gi->ct->lim.lam*4, fid ) 
+            if( pj_ctx_fread( ctx, row_buf, sizeof(float),
+                              gi->ct->lim.lam*4, fid )
                 != 4 * gi->ct->lim.lam )
             {
                 pj_dalloc( row_buf );
-                pj_dalloc( gi->ct->cvs );
-                gi->ct->cvs = NULL;
-                pj_errno = -38;
+                pj_dalloc( ct_tmp.cvs );
+                pj_ctx_set_errno( ctx, -38 );
+                pj_release_lock();
                 return 0;
             }
 
             if( !IS_LSB )
-                swap_words( (unsigned char *) row_buf, 4, 
+                swap_words( (unsigned char *) row_buf, 4,
                             gi->ct->lim.lam*4 );
 
             /* convert seconds to radians */
@@ -281,7 +323,7 @@ int pj_gridinfo_load( PJ_GRIDINFO *gi )
 
             for( i = 0; i < gi->ct->lim.lam; i++ )
             {
-                cvs = gi->ct->cvs + (row) * gi->ct->lim.lam
+                cvs = ct_tmp.cvs + (row) * gi->ct->lim.lam
                     + (gi->ct->lim.lam - i - 1);
 
                 cvs->phi = *(diff_seconds++) * ((PI/180.0) / 3600.0);
@@ -292,25 +334,94 @@ int pj_gridinfo_load( PJ_GRIDINFO *gi )
 
         pj_dalloc( row_buf );
 
-        fclose( fid );
+        pj_ctx_fclose( ctx, fid );
 
+        gi->ct->cvs = ct_tmp.cvs;
+
+        pj_release_lock();
+        return 1;
+    }
+
+/* -------------------------------------------------------------------- */
+/*      GTX format.                                                     */
+/* -------------------------------------------------------------------- */
+    else if( strcmp(gi->format,"gtx") == 0 )
+    {
+        int   words = gi->ct->lim.lam * gi->ct->lim.phi;
+        PAFile fid;
+
+        fid = pj_open_lib( ctx, gi->filename, "rb" );
+
+        if( fid == NULL )
+        {
+            pj_ctx_set_errno( ctx, -38 );
+            pj_release_lock();
+            return 0;
+        }
+
+        pj_ctx_fseek( ctx, fid, gi->grid_offset, SEEK_SET );
+
+        ct_tmp.cvs = (FLP *) pj_malloc(words*sizeof(float));
+        if( ct_tmp.cvs == NULL )
+        {
+            pj_ctx_set_errno( ctx, -38 );
+            pj_release_lock();
+            return 0;
+        }
+
+        if( pj_ctx_fread( ctx, ct_tmp.cvs, sizeof(float), words, fid )
+            != words )
+        {
+            pj_dalloc( ct_tmp.cvs );
+            pj_release_lock();
+            return 0;
+        }
+
+        if( IS_LSB )
+            swap_words( (unsigned char *) ct_tmp.cvs, 4, words );
+
+        pj_ctx_fclose( ctx, fid );
+        gi->ct->cvs = ct_tmp.cvs;
+        pj_release_lock();
         return 1;
     }
 
     else
     {
+        pj_release_lock();
         return 0;
     }
 }
 
+/************************************************************************/
+/*                       pj_gridinfo_init_ntv2()                        */
+/*                                                                      */
+/*      Seek a parent grid file by name from a grid list                */
+/************************************************************************/
+
+static PJ_GRIDINFO* pj_gridinfo_parent( PJ_GRIDINFO *gilist, 
+        const char *name, int length )
+{
+    while( gilist )
+    {
+        if( strncmp(gilist->ct->id,name,length) == 0 ) return gilist;
+        if( gilist->child )
+        {
+            PJ_GRIDINFO *parent=pj_gridinfo_parent( gilist->child, name, length );
+            if( parent ) return parent;
+        }
+        gilist=gilist->next;
+    }
+    return gilist;
+}
+
 /************************************************************************/
 /*                       pj_gridinfo_init_ntv2()                        */
 /*                                                                      */
 /*      Load a ntv2 (.gsb) file.                                        */
 /************************************************************************/
 
-static int pj_gridinfo_init_ntv2( FILE *fid, PJ_GRIDINFO *gilist )
-
+static int pj_gridinfo_init_ntv2( projCtx ctx, PAFile fid, PJ_GRIDINFO *gilist )
 {
     unsigned char header[11*16];
     int num_subfiles, subfile;
@@ -319,18 +430,18 @@ static int pj_gridinfo_init_ntv2( FILE *fid, PJ_GRIDINFO *gilist )
     assert( sizeof(double) == 8 );
     if( sizeof(int) != 4 || sizeof(double) != 8 )
     {
-        fprintf( stderr, 
-                 "basic types of inappropraiate size in pj_gridinfo_init_ntv2()\n" );
-        pj_errno = -38;
+        pj_log( ctx, PJ_LOG_ERROR,
+             "basic types of inappropraiate size in pj_gridinfo_init_ntv2()" );
+        pj_ctx_set_errno( ctx, -38 );
         return 0;
     }
 
 /* -------------------------------------------------------------------- */
 /*      Read the overview header.                                       */
 /* -------------------------------------------------------------------- */
-    if( fread( header, sizeof(header), 1, fid ) != 1 )
+    if( pj_ctx_fread( ctx, header, sizeof(header), 1, fid ) != 1 )
     {
-        pj_errno = -38;
+        pj_ctx_set_errno( ctx, -38 );
         return 0;
     }
 
@@ -366,18 +477,18 @@ static int pj_gridinfo_init_ntv2( FILE *fid, PJ_GRIDINFO *gilist )
 /* -------------------------------------------------------------------- */
 /*      Read header.                                                    */
 /* -------------------------------------------------------------------- */
-        if( fread( header, sizeof(header), 1, fid ) != 1 )
+        if( pj_ctx_fread( ctx, header, sizeof(header), 1, fid ) != 1 )
         {
-            pj_errno = -38;
+            pj_ctx_set_errno( ctx, -38 );
             return 0;
         }
 
-        if( strncmp(header,"SUB_NAME",8) != 0 )
+        if( strncmp((const char *) header,"SUB_NAME",8) != 0 )
         {
-            pj_errno = -38;
+            pj_ctx_set_errno( ctx, -38 );
             return 0;
         }
-        
+
 /* -------------------------------------------------------------------- */
 /*      Byte swap interesting fields if needed.                         */
 /* -------------------------------------------------------------------- */
@@ -391,12 +502,12 @@ static int pj_gridinfo_init_ntv2( FILE *fid, PJ_GRIDINFO *gilist )
             swap_words( header+8+16*9, 8, 1 );
             swap_words( header+8+16*10, 4, 1 );
         }
-        
+
 /* -------------------------------------------------------------------- */
 /*      Initialize a corresponding "ct" structure.                      */
 /* -------------------------------------------------------------------- */
         ct = (struct CTABLE *) pj_malloc(sizeof(struct CTABLE));
-        strncpy( ct->id, header + 8, 8 );
+        strncpy( ct->id, (const char *) header + 8, 8 );
         ct->id[8] = '\0';
 
         ct->ll.lam = - *((double *) (header+7*16+8)); /* W_LONG */
@@ -411,13 +522,12 @@ static int pj_gridinfo_init_ntv2( FILE *fid, PJ_GRIDINFO *gilist )
         ct->lim.lam = (int) (fabs(ur.lam-ct->ll.lam)/ct->del.lam + 0.5) + 1;
         ct->lim.phi = (int) (fabs(ur.phi-ct->ll.phi)/ct->del.phi + 0.5) + 1;
 
-        if( getenv("PROJ_DEBUG") != NULL )
-            fprintf( stderr, 
-                     "NTv2 %s %dx%d: LL=(%.9g,%.9g) UR=(%.9g,%.9g)\n",
-                     ct->id, 
-                     ct->lim.lam, ct->lim.phi,
-                     ct->ll.lam/3600.0, ct->ll.phi/3600.0,
-                     ur.lam/3600.0, ur.phi/3600.0 );
+        pj_log( ctx, PJ_LOG_DEBUG_MINOR,
+                "NTv2 %s %dx%d: LL=(%.9g,%.9g) UR=(%.9g,%.9g)\n",
+                ct->id,
+                ct->lim.lam, ct->lim.phi,
+                ct->ll.lam/3600.0, ct->ll.phi/3600.0,
+                ur.lam/3600.0, ur.phi/3600.0 );
 
         ct->ll.lam *= DEG_TO_RAD/3600.0;
         ct->ll.phi *= DEG_TO_RAD/3600.0;
@@ -427,11 +537,11 @@ static int pj_gridinfo_init_ntv2( FILE *fid, PJ_GRIDINFO *gilist )
         memcpy( &gs_count, header + 8 + 16*10, 4 );
         if( gs_count != ct->lim.lam * ct->lim.phi )
         {
-            fprintf( stderr, 
-                     "GS_COUNT(%d) does not match expected cells (%dx%d=%d)\n",
-                     gs_count, ct->lim.lam, ct->lim.phi, 
-                     ct->lim.lam * ct->lim.phi );
-            pj_errno = -38;
+            pj_log( ctx, PJ_LOG_ERROR,
+                    "GS_COUNT(%d) does not match expected cells (%dx%d=%d)\n",
+                    gs_count, ct->lim.lam, ct->lim.phi,
+                    ct->lim.lam * ct->lim.phi );
+            pj_ctx_set_errno( ctx, -38 );
             return 0;
         }
 
@@ -447,7 +557,7 @@ static int pj_gridinfo_init_ntv2( FILE *fid, PJ_GRIDINFO *gilist )
         {
             gi = (PJ_GRIDINFO *) pj_malloc(sizeof(PJ_GRIDINFO));
             memset( gi, 0, sizeof(PJ_GRIDINFO) );
-    
+
             gi->gridname = strdup( gilist->gridname );
             gi->filename = strdup( gilist->filename );
             gi->next = NULL;
@@ -455,12 +565,12 @@ static int pj_gridinfo_init_ntv2( FILE *fid, PJ_GRIDINFO *gilist )
 
         gi->ct = ct;
         gi->format = "ntv2";
-        gi->grid_offset = ftell( fid );
+        gi->grid_offset = pj_ctx_ftell( ctx, fid );
 
 /* -------------------------------------------------------------------- */
 /*      Attach to the correct list or sublist.                          */
 /* -------------------------------------------------------------------- */
-        if( strncmp(header+24,"NONE",4) == 0 )
+        if( strncmp((const char *)header+24,"NONE",4) == 0 )
         {
             if( gi != gilist )
             {
@@ -474,36 +584,37 @@ static int pj_gridinfo_init_ntv2( FILE *fid, PJ_GRIDINFO *gilist )
         else
         {
             PJ_GRIDINFO *lnk;
-            PJ_GRIDINFO *gp = gilist;
-            
-            while( gp != NULL && strncmp(gp->ct->id,header+24,8) != 0 )
-                gp = gp->next;
+            PJ_GRIDINFO *gp = pj_gridinfo_parent(gilist,
+                                                 (const char*)header+24,8);
 
             if( gp == NULL )
             {
-                if( getenv("PROJ_DEBUG") != NULL )
-                    fprintf( stderr, "pj_gridinfo_init_ntv2(): "
-                             "failed to find parent %8.8s for %.\n", 
-                             header+24, gi->ct->id );
+                pj_log( ctx, PJ_LOG_ERROR,
+                        "pj_gridinfo_init_ntv2(): "
+                        "failed to find parent %8.8s for %s.\n",
+                        (const char *) header+24, gi->ct->id );
 
-                for( lnk = gp; lnk->next != NULL; lnk = lnk->next ) {}
+                for( lnk = gilist; lnk->next != NULL; lnk = lnk->next ) {}
                 lnk->next = gi;
             }
-            else if( gp->child == NULL )
-            {
-                gp->child = gi;
-            }
             else
             {
-                for( lnk = gp->child; lnk->next != NULL; lnk = lnk->next ) {}
-                lnk->next = gi;
+                if( gp->child == NULL )
+                {
+                    gp->child = gi;
+                }
+                else
+                {
+                    for( lnk = gp->child; lnk->next != NULL; lnk = lnk->next ) {}
+                    lnk->next = gi;
+                }
             }
         }
 
 /* -------------------------------------------------------------------- */
 /*      Seek past the data.                                             */
 /* -------------------------------------------------------------------- */
-        fseek( fid, gs_count * 16, SEEK_CUR );
+        pj_ctx_fseek( ctx, fid, gs_count * 16, SEEK_CUR );
     }
 
     return 1;
@@ -515,29 +626,29 @@ static int pj_gridinfo_init_ntv2( FILE *fid, PJ_GRIDINFO *gilist )
 /*      Load an NTv1 style Canadian grid shift file.                    */
 /************************************************************************/
 
-static int pj_gridinfo_init_ntv1( FILE * fid, PJ_GRIDINFO *gi )
+static int pj_gridinfo_init_ntv1( projCtx ctx, PAFile fid, PJ_GRIDINFO *gi )
 
 {
     unsigned char header[176];
     struct CTABLE *ct;
     LP		ur;
-    
+
     assert( sizeof(int) == 4 );
     assert( sizeof(double) == 8 );
     if( sizeof(int) != 4 || sizeof(double) != 8 )
     {
-        fprintf( stderr, 
-                 "basic types of inappropraiate size in nad_load_ntv1()\n" );
-        pj_errno = -38;
+        pj_log( ctx, PJ_LOG_ERROR,
+                 "basic types of inappropraiate size in nad_load_ntv1()" );
+        pj_ctx_set_errno( ctx, -38 );
         return 0;
     }
 
 /* -------------------------------------------------------------------- */
 /*      Read the header.                                                */
 /* -------------------------------------------------------------------- */
-    if( fread( header, sizeof(header), 1, fid ) != 1 )
+    if( pj_ctx_fread( ctx, header, sizeof(header), 1, fid ) != 1 )
     {
-        pj_errno = -38;
+        pj_ctx_set_errno( ctx, -38 );
         return 0;
     }
 
@@ -557,8 +668,9 @@ static int pj_gridinfo_init_ntv1( FILE * fid, PJ_GRIDINFO *gi )
 
     if( *((int *) (header+8)) != 12 )
     {
-        pj_errno = -38;
-        printf("NTv1 grid shift file has wrong record count, corrupt?\n");
+        pj_log( ctx, PJ_LOG_ERROR,
+                "NTv1 grid shift file has wrong record count, corrupt?" );
+        pj_ctx_set_errno( ctx, -38 );
         return 0;
     }
 
@@ -577,11 +689,10 @@ static int pj_gridinfo_init_ntv1( FILE * fid, PJ_GRIDINFO *gi )
     ct->lim.lam = (int) (fabs(ur.lam-ct->ll.lam)/ct->del.lam + 0.5) + 1;
     ct->lim.phi = (int) (fabs(ur.phi-ct->ll.phi)/ct->del.phi + 0.5) + 1;
 
-    if( getenv("PROJ_DEBUG") != NULL )
-        fprintf( stderr, 
-                 "NTv1 %dx%d: LL=(%.9g,%.9g) UR=(%.9g,%.9g)\n",
-                 ct->lim.lam, ct->lim.phi,
-                 ct->ll.lam, ct->ll.phi, ur.lam, ur.phi );
+    pj_log( ctx, PJ_LOG_DEBUG_MINOR,
+            "NTv1 %dx%d: LL=(%.9g,%.9g) UR=(%.9g,%.9g)",
+            ct->lim.lam, ct->lim.phi,
+            ct->ll.lam, ct->ll.phi, ur.lam, ur.phi );
 
     ct->ll.lam *= DEG_TO_RAD;
     ct->ll.phi *= DEG_TO_RAD;
@@ -590,12 +701,115 @@ static int pj_gridinfo_init_ntv1( FILE * fid, PJ_GRIDINFO *gi )
     ct->cvs = NULL;
 
     gi->ct = ct;
-    gi->grid_offset = ftell( fid );
+    gi->grid_offset = pj_ctx_ftell( ctx, fid );
     gi->format = "ntv1";
 
     return 1;
 }
 
+/************************************************************************/
+/*                       pj_gridinfo_init_gtx()                         */
+/*                                                                      */
+/*      Load a NOAA .gtx vertical datum shift file.                     */
+/************************************************************************/
+
+static int pj_gridinfo_init_gtx( projCtx ctx, PAFile fid, PJ_GRIDINFO *gi )
+
+{
+    unsigned char header[40];
+    struct CTABLE *ct;
+    double      xorigin,yorigin,xstep,ystep;
+    int         rows, columns;
+
+    assert( sizeof(int) == 4 );
+    assert( sizeof(double) == 8 );
+    if( sizeof(int) != 4 || sizeof(double) != 8 )
+    {
+        pj_log( ctx, PJ_LOG_ERROR,
+                "basic types of inappropraiate size in nad_load_gtx()" );
+        pj_ctx_set_errno( ctx, -38 );
+        return 0;
+    }
+
+/* -------------------------------------------------------------------- */
+/*      Read the header.                                                */
+/* -------------------------------------------------------------------- */
+    if( pj_ctx_fread( ctx, header, sizeof(header), 1, fid ) != 1 )
+    {
+        pj_ctx_set_errno( ctx, -38 );
+        return 0;
+    }
+
+/* -------------------------------------------------------------------- */
+/*      Regularize fields of interest and extract.                      */
+/* -------------------------------------------------------------------- */
+    if( IS_LSB )
+    {
+        swap_words( header+0, 8, 4 );
+        swap_words( header+32, 4, 2 );
+    }
+
+    memcpy( &yorigin, header+0, 8 );
+    memcpy( &xorigin, header+8, 8 );
+    memcpy( &ystep, header+16, 8 );
+    memcpy( &xstep, header+24, 8 );
+
+    memcpy( &rows, header+32, 4 );
+    memcpy( &columns, header+36, 4 );
+
+    if( xorigin < -360 || xorigin > 360
+        || yorigin < -90 || yorigin > 90 )
+    {
+        pj_log( ctx, PJ_LOG_ERROR,
+                "gtx file header has invalid extents, corrupt?");
+        pj_ctx_set_errno( ctx, -38 );
+        return 0;
+    }
+
+/* -------------------------------------------------------------------- */
+/*      Fill in CTABLE structure.                                       */
+/* -------------------------------------------------------------------- */
+    ct = (struct CTABLE *) pj_malloc(sizeof(struct CTABLE));
+    strcpy( ct->id, "GTX Vertical Grid Shift File" );
+
+    ct->ll.lam = xorigin;
+    ct->ll.phi = yorigin;
+    ct->del.lam = xstep;
+    ct->del.phi = ystep;
+    ct->lim.lam = columns;
+    ct->lim.phi = rows;
+
+    /* some GTX files come in 0-360 and we shift them back into the
+       expected -180 to 180 range if possible.  This does not solve
+       problems with grids spanning the dateline. */
+    if( ct->ll.lam >= 180.0 )
+        ct->ll.lam -= 360.0;
+
+    if( ct->ll.lam >= 0.0 && ct->ll.lam + ct->del.lam * ct->lim.lam > 180.0 )
+    {
+        pj_log( ctx, PJ_LOG_DEBUG_MAJOR,
+                "This GTX spans the dateline!  This will cause problems." );
+    }
+
+    pj_log( ctx, PJ_LOG_DEBUG_MINOR,
+            "GTX %dx%d: LL=(%.9g,%.9g) UR=(%.9g,%.9g)",
+            ct->lim.lam, ct->lim.phi,
+            ct->ll.lam, ct->ll.phi,
+            ct->ll.lam + (columns-1)*xstep, ct->ll.phi + (rows-1)*ystep);
+
+    ct->ll.lam *= DEG_TO_RAD;
+    ct->ll.phi *= DEG_TO_RAD;
+    ct->del.lam *= DEG_TO_RAD;
+    ct->del.phi *= DEG_TO_RAD;
+    ct->cvs = NULL;
+
+    gi->ct = ct;
+    gi->grid_offset = 40;
+    gi->format = "gtx";
+
+    return 1;
+}
+
 /************************************************************************/
 /*                          pj_gridinfo_init()                          */
 /*                                                                      */
@@ -605,15 +819,16 @@ static int pj_gridinfo_init_ntv1( FILE * fid, PJ_GRIDINFO *gi )
 /*      applications.                                                   */
 /************************************************************************/
 
-PJ_GRIDINFO *pj_gridinfo_init( const char *gridname )
+PJ_GRIDINFO *pj_gridinfo_init( projCtx ctx, const char *gridname )
 
 {
     char 	fname[MAX_PATH_FILENAME+1];
     PJ_GRIDINFO *gilist;
-    FILE 	*fp;
+    PAFile 	fp;
     char	header[160];
 
     errno = pj_errno = 0;
+    ctx->last_errno = 0;
 
 /* -------------------------------------------------------------------- */
 /*      Initialize a GRIDINFO with stub info we would use if it         */
@@ -621,7 +836,7 @@ PJ_GRIDINFO *pj_gridinfo_init( const char *gridname )
 /* -------------------------------------------------------------------- */
     gilist = (PJ_GRIDINFO *) pj_malloc(sizeof(PJ_GRIDINFO));
     memset( gilist, 0, sizeof(PJ_GRIDINFO) );
-    
+
     gilist->gridname = strdup( gridname );
     gilist->filename = NULL;
     gilist->format = "missing";
@@ -633,59 +848,87 @@ PJ_GRIDINFO *pj_gridinfo_init( const char *gridname )
 /*      Open the file using the usual search rules.                     */
 /* -------------------------------------------------------------------- */
     strcpy(fname, gridname);
-    if (!(fp = pj_open_lib(fname, "rb"))) {
-        pj_errno = errno;
+    if (!(fp = pj_open_lib(ctx, fname, "rb"))) {
+        ctx->last_errno = 0; /* don't treat as a persistent error */
         return gilist;
     }
 
     gilist->filename = strdup(fname);
-    
+
 /* -------------------------------------------------------------------- */
 /*      Load a header, to determine the file type.                      */
 /* -------------------------------------------------------------------- */
-    if( fread( header, sizeof(header), 1, fp ) != 1 )
+    if( pj_ctx_fread( ctx, header, sizeof(header), 1, fp ) != 1 )
     {
-        fclose( fp );
-        pj_errno = -38;
+        pj_ctx_fclose( ctx, fp );
+        pj_ctx_set_errno( ctx, -38 );
         return gilist;
     }
 
-    fseek( fp, SEEK_SET, 0 );
+    pj_ctx_fseek( ctx, fp, SEEK_SET, 0 );
 
 /* -------------------------------------------------------------------- */
 /*      Determine file type.                                            */
 /* -------------------------------------------------------------------- */
-    if( strncmp(header + 0, "HEADER", 6) == 0 
-        && strncmp(header + 96, "W GRID", 6) == 0 
+    if( strncmp(header + 0, "HEADER", 6) == 0
+        && strncmp(header + 96, "W GRID", 6) == 0
         && strncmp(header + 144, "TO      NAD83   ", 16) == 0 )
     {
-        pj_gridinfo_init_ntv1( fp, gilist );
+        pj_gridinfo_init_ntv1( ctx, fp, gilist );
     }
-    
-    else if( strncmp(header + 0, "NUM_OREC", 8) == 0 
+
+    else if( strncmp(header + 0, "NUM_OREC", 8) == 0
              && strncmp(header + 48, "GS_TYPE", 7) == 0 )
     {
-        pj_gridinfo_init_ntv2( fp, gilist );
+        pj_gridinfo_init_ntv2( ctx, fp, gilist );
     }
-    
-    else
-    {
-        struct CTABLE *ct = nad_ctable_init( fp );
 
-        gilist->format = "ctable";
+    else if( strlen(gridname) > 4
+             && (strcmp(gridname+strlen(gridname)-3,"gtx") == 0
+                 || strcmp(gridname+strlen(gridname)-3,"GTX") == 0) )
+    {
+        pj_gridinfo_init_gtx( ctx, fp, gilist );
+    }
+
+    else if( strncmp(header + 0,"CTABLE V2",9) == 0 )
+    {
+        struct CTABLE *ct = nad_ctable2_init( ctx, fp );
+
+        gilist->format = "ctable2";
         gilist->ct = ct;
 
-        if( getenv("PROJ_DEBUG") != NULL )
-            fprintf( stderr, 
-                     "Ctable %s %dx%d: LL=(%.9g,%.9g) UR=(%.9g,%.9g)\n",
-                     ct->id, 
-                     ct->lim.lam, ct->lim.phi,
-                     ct->ll.lam * RAD_TO_DEG, ct->ll.phi * RAD_TO_DEG,
-                     (ct->ll.lam + ct->lim.lam*ct->del.lam) * RAD_TO_DEG, 
-                     (ct->ll.phi + ct->lim.phi*ct->del.lam) * RAD_TO_DEG );
+        pj_log( ctx, PJ_LOG_DEBUG_MAJOR,
+                "Ctable2 %s %dx%d: LL=(%.9g,%.9g) UR=(%.9g,%.9g)\n",
+                ct->id,
+                ct->lim.lam, ct->lim.phi,
+                ct->ll.lam * RAD_TO_DEG, ct->ll.phi * RAD_TO_DEG,
+                (ct->ll.lam + (ct->lim.lam-1)*ct->del.lam) * RAD_TO_DEG,
+                (ct->ll.phi + (ct->lim.phi-1)*ct->del.phi) * RAD_TO_DEG );
     }
 
-    fclose(fp);
+    else
+    {
+        struct CTABLE *ct = nad_ctable_init( ctx, fp );
+        if (ct == NULL)
+        {
+            pj_log( ctx, PJ_LOG_DEBUG_MAJOR,
+                    "CTABLE ct is NULL.");
+        } else
+        {
+            gilist->format = "ctable";
+            gilist->ct = ct;
+
+            pj_log( ctx, PJ_LOG_DEBUG_MAJOR,
+                    "Ctable %s %dx%d: LL=(%.9g,%.9g) UR=(%.9g,%.9g)\n",
+                    ct->id,
+                    ct->lim.lam, ct->lim.phi,
+                    ct->ll.lam * RAD_TO_DEG, ct->ll.phi * RAD_TO_DEG,
+                    (ct->ll.lam + (ct->lim.lam-1)*ct->del.lam) * RAD_TO_DEG,
+                    (ct->ll.phi + (ct->lim.phi-1)*ct->del.phi) * RAD_TO_DEG );
+        }
+    }
+
+    pj_ctx_fclose(ctx, fp);
 
     return gilist;
 }
