@@ -1,5 +1,5 @@
 /******************************************************************************
- * $Id: pj_gridlist.c,v 1.3 2003/03/18 16:26:58 warmerda Exp $
+ * $Id$
  *
  * Project:  PROJ.4
  * Purpose:  Code to manage the list of currently loaded (cached) PJ_GRIDINFOs
@@ -26,36 +26,26 @@
  * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
  * FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
  * DEALINGS IN THE SOFTWARE.
- ******************************************************************************
- *
- * $Log: pj_gridlist.c,v $
- * Revision 1.3  2003/03/18 16:26:58  warmerda
- * clear error if missing file is not required
- *
- * Revision 1.2  2003/03/17 19:45:47  warmerda
- * support '@' marker for optional grids
- *
- * Revision 1.1  2003/03/15 06:01:18  warmerda
- * New
- *
- */
+ *****************************************************************************/
 
 #define PJ_LIB__
 
 #include <projects.h>
 #include <string.h>
 #include <math.h>
-#include <assert.h>
+
+#ifdef _WIN32_WCE
+/* assert.h includes all Windows API headers and causes 'LP' name clash.
+ * Here assert we disable assert() for Windows CE.
+ * TODO - mloskot: re-implement porting friendly assert
+ */
+# define assert(exp)	((void)0)
+#else
+# include <assert.h>
+#endif /* _WIN32_WCE */
 
 static PJ_GRIDINFO *grid_list = NULL;
 
-/* used only by pj_load_nadgrids() and pj_deallocate_grids() */
-
-static int           last_nadgrids_max = 0;
-static int           last_nadgrids_count = 0;
-static PJ_GRIDINFO **last_nadgrids_list = NULL;
-static char         *last_nadgrids = NULL;
-
 /************************************************************************/
 /*                        pj_deallocate_grids()                         */
 /*                                                                      */
@@ -71,19 +61,7 @@ void pj_deallocate_grids()
         grid_list = grid_list->next;
         item->next = NULL;
 
-        pj_gridinfo_free( item );
-    }
-
-    if( last_nadgrids != NULL )
-    {
-        pj_dalloc( last_nadgrids );
-        last_nadgrids = NULL;
-
-        pj_dalloc( last_nadgrids_list );
-        last_nadgrids_list = NULL;
-
-        last_nadgrids_count = 0;
-        last_nadgrids_max = 0;
+        pj_gridinfo_free( pj_get_default_ctx(), item );
     }
 }
 
@@ -94,10 +72,14 @@ void pj_deallocate_grids()
 /*      last_nadgrids_list.                                             */
 /************************************************************************/
 
-static int pj_gridlist_merge_gridfile( const char *gridname )
+static int pj_gridlist_merge_gridfile( projCtx ctx, 
+                                       const char *gridname,
+                                       PJ_GRIDINFO ***p_gridlist,
+                                       int *p_gridcount, 
+                                       int *p_gridmax )
 
 {
-    int i, got_match=0;
+    int got_match=0;
     PJ_GRIDINFO *this_grid, *tail = NULL;
 
 /* -------------------------------------------------------------------- */
@@ -116,26 +98,26 @@ static int pj_gridlist_merge_gridfile( const char *gridname )
                 return 0;
 
             /* do we need to grow the list? */
-            if( last_nadgrids_count >= last_nadgrids_max - 2 )
+            if( *p_gridcount >= *p_gridmax - 2 )
             {
                 PJ_GRIDINFO **new_list;
-                int new_max = last_nadgrids_max + 20;
+                int new_max = *p_gridmax + 20;
 
                 new_list = (PJ_GRIDINFO **) pj_malloc(sizeof(void*) * new_max);
-                if( last_nadgrids_list != NULL )
+                if( *p_gridlist != NULL )
                 {
-                    memcpy( new_list, last_nadgrids_list, 
-                            sizeof(void*) * last_nadgrids_max );
-                    pj_dalloc( last_nadgrids_list );
+                    memcpy( new_list, *p_gridlist,
+                            sizeof(void*) * (*p_gridmax) );
+                    pj_dalloc( *p_gridlist );
                 }
 
-                last_nadgrids_list = new_list;
-                last_nadgrids_max = new_max;
+                *p_gridlist = new_list;
+                *p_gridmax = new_max;
             }
 
             /* add to the list */
-            last_nadgrids_list[last_nadgrids_count++] = this_grid;
-            last_nadgrids_list[last_nadgrids_count] = NULL;
+            (*p_gridlist)[(*p_gridcount)++] = this_grid;
+            (*p_gridlist)[*p_gridcount] = NULL;
         }
 
         tail = this_grid;
@@ -147,7 +129,7 @@ static int pj_gridlist_merge_gridfile( const char *gridname )
 /* -------------------------------------------------------------------- */
 /*      Try to load the named grid.                                     */
 /* -------------------------------------------------------------------- */
-    this_grid = pj_gridinfo_init( gridname );
+    this_grid = pj_gridinfo_init( ctx, gridname );
 
     if( this_grid == NULL )
     {
@@ -164,7 +146,8 @@ static int pj_gridlist_merge_gridfile( const char *gridname )
 /* -------------------------------------------------------------------- */
 /*      Recurse to add the grid now that it is loaded.                  */
 /* -------------------------------------------------------------------- */
-    return pj_gridlist_merge_gridfile( gridname );
+    return pj_gridlist_merge_gridfile( ctx, gridname, p_gridlist, 
+                                       p_gridcount, p_gridmax );
 }
 
 /************************************************************************/
@@ -177,33 +160,18 @@ static int pj_gridlist_merge_gridfile( const char *gridname )
 /*      the cost of building the list of tables each time.              */
 /************************************************************************/
 
-PJ_GRIDINFO **pj_gridlist_from_nadgrids( const char *nadgrids, int *grid_count)
+PJ_GRIDINFO **pj_gridlist_from_nadgrids( projCtx ctx, const char *nadgrids, 
+                                         int *grid_count)
 
 {
     const char *s;
+    PJ_GRIDINFO **gridlist = NULL;
+    int grid_max = 0;
 
     pj_errno = 0;
     *grid_count = 0;
 
-    if( last_nadgrids != NULL 
-        && strcmp(nadgrids,last_nadgrids) == 0 )
-    {
-        *grid_count = last_nadgrids_count;
-        return last_nadgrids_list;
-    }
-
-/* -------------------------------------------------------------------- */
-/*      Free old one, if any, and make space for new list.              */
-/* -------------------------------------------------------------------- */
-    if( last_nadgrids != NULL )
-    {
-        pj_dalloc(last_nadgrids);
-    }
-    
-    last_nadgrids = (char *) pj_malloc(strlen(nadgrids)+1);
-    strcpy( last_nadgrids, nadgrids );
-
-    last_nadgrids_count = 0;
+    pj_acquire_lock();
 
 /* -------------------------------------------------------------------- */
 /*      Loop processing names out of nadgrids one at a time.            */
@@ -224,9 +192,10 @@ PJ_GRIDINFO **pj_gridlist_from_nadgrids( const char *nadgrids, int *grid_count)
              s[end_char] != '\0' && s[end_char] != ','; 
              end_char++ ) {}
 
-        if( end_char > sizeof(name) )
+        if( end_char >= sizeof(name) )
         {
-            pj_errno = -38;
+            pj_ctx_set_errno( ctx, -38 );
+            pj_release_lock();
             return NULL;
         }
         
@@ -237,20 +206,19 @@ PJ_GRIDINFO **pj_gridlist_from_nadgrids( const char *nadgrids, int *grid_count)
         if( *s == ',' )
             s++;
 
-        if( !pj_gridlist_merge_gridfile( name ) && required )
+        if( !pj_gridlist_merge_gridfile( ctx, name, &gridlist, grid_count, 
+                                         &grid_max) 
+            && required )
         {
-            pj_errno = -38;
+            pj_ctx_set_errno( ctx, -38 );
+            pj_release_lock();
             return NULL;
         }
         else
             pj_errno = 0;
     }
 
-    if( last_nadgrids_count > 0 )
-    {
-        *grid_count = last_nadgrids_count;
-        return last_nadgrids_list;
-    }
-    else
-        return NULL;
+    pj_release_lock();
+
+    return gridlist;
 }
